# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/galacteek.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_GalacteekWindow(object):
    def setupUi(self, GalacteekWindow):
        GalacteekWindow.setObjectName("GalacteekWindow")
        GalacteekWindow.resize(1038, 501)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(GalacteekWindow.sizePolicy().hasHeightForWidth())
        GalacteekWindow.setSizePolicy(sizePolicy)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/share/icons/galacteek.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        GalacteekWindow.setWindowIcon(icon)
        self.centralwidget = QtWidgets.QWidget(GalacteekWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.tabWidget = QtWidgets.QTabWidget(self.centralwidget)
        self.tabWidget.setObjectName("tabWidget")
        self.tabBlank = QtWidgets.QWidget()
        self.tabBlank.setObjectName("tabBlank")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.tabBlank)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.tabWidget.addTab(self.tabBlank, "")
        self.verticalLayout.addWidget(self.tabWidget)
        self.verticalLayout_3.addLayout(self.verticalLayout)
        GalacteekWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(GalacteekWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1038, 20))
        self.menubar.setDefaultUp(True)
        self.menubar.setObjectName("menubar")
        GalacteekWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(GalacteekWindow)
        self.statusbar.setObjectName("statusbar")
        GalacteekWindow.setStatusBar(self.statusbar)
        self.actionQuit = QtWidgets.QAction(GalacteekWindow)
        self.actionQuit.setPriority(QtWidgets.QAction.LowPriority)
        self.actionQuit.setObjectName("actionQuit")
        self.actionAddProfile = QtWidgets.QAction(GalacteekWindow)
        self.actionAddProfile.setObjectName("actionAddProfile")
        self.actionShowPeersInformation = QtWidgets.QAction(GalacteekWindow)
        self.actionShowPeersInformation.setObjectName("actionShowPeersInformation")
        self.actionShowIpfsRefs = QtWidgets.QAction(GalacteekWindow)
        self.actionShowIpfsRefs.setObjectName("actionShowIpfsRefs")
        self.actionSettings = QtWidgets.QAction(GalacteekWindow)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/share/icons/settings.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionSettings.setIcon(icon1)
        self.actionSettings.setObjectName("actionSettings")
        self.actionAboutGalacteek = QtWidgets.QAction(GalacteekWindow)
        self.actionAboutGalacteek.setObjectName("actionAboutGalacteek")
        self.actionCloseAllTabs = QtWidgets.QAction(GalacteekWindow)
        self.actionCloseAllTabs.setPriority(QtWidgets.QAction.HighPriority)
        self.actionCloseAllTabs.setObjectName("actionCloseAllTabs")
        self.actionNew_Profile = QtWidgets.QAction(GalacteekWindow)
        self.actionNew_Profile.setObjectName("actionNew_Profile")
        self.actionDonate = QtWidgets.QAction(GalacteekWindow)
        self.actionDonate.setObjectName("actionDonate")
        self.actionEvent_log = QtWidgets.QAction(GalacteekWindow)
        self.actionEvent_log.setObjectName("actionEvent_log")
        self.actionBrowse = QtWidgets.QAction(GalacteekWindow)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/share/icons/ipfs-logo-128-white-outline.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionBrowse.setIcon(icon2)
        self.actionBrowse.setObjectName("actionBrowse")
        self.actionExitApp = QtWidgets.QAction(GalacteekWindow)
        self.actionExitApp.setObjectName("actionExitApp")

        self.retranslateUi(GalacteekWindow)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(GalacteekWindow)

    def retranslateUi(self, GalacteekWindow):
        _translate = QtCore.QCoreApplication.translate
        GalacteekWindow.setWindowTitle(_translate("GalacteekWindow", "Galacteek"))
        self.actionQuit.setText(_translate("GalacteekWindow", "Quit"))
        self.actionQuit.setShortcut(_translate("GalacteekWindow", "Ctrl+Q"))
        self.actionAddProfile.setText(_translate("GalacteekWindow", "Add Profile"))
        self.actionShowPeersInformation.setText(_translate("GalacteekWindow", "Show peers information"))
        self.actionShowIpfsRefs.setText(_translate("GalacteekWindow", "Show refs"))
        self.actionSettings.setText(_translate("GalacteekWindow", "Settings"))
        self.actionAboutGalacteek.setText(_translate("GalacteekWindow", "About"))
        self.actionCloseAllTabs.setText(_translate("GalacteekWindow", "Close all tabs"))
        self.actionNew_Profile.setText(_translate("GalacteekWindow", "New Profile"))
        self.actionDonate.setText(_translate("GalacteekWindow", "Donate"))
        self.actionEvent_log.setText(_translate("GalacteekWindow", "Event log"))
        self.actionBrowse.setText(_translate("GalacteekWindow", "Browse"))
        self.actionExitApp.setText(_translate("GalacteekWindow", "Quit"))

from . import galacteek_rc
