# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/browsertab.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_BrowserTabForm(object):
    def setupUi(self, BrowserTabForm):
        BrowserTabForm.setObjectName("BrowserTabForm")
        BrowserTabForm.resize(843, 444)
        self.verticalLayout_18 = QtWidgets.QVBoxLayout(BrowserTabForm)
        self.verticalLayout_18.setObjectName("verticalLayout_18")
        self.verticalLayout_17 = QtWidgets.QVBoxLayout()
        self.verticalLayout_17.setContentsMargins(-1, -1, -1, 0)
        self.verticalLayout_17.setObjectName("verticalLayout_17")
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setSizeConstraint(QtWidgets.QLayout.SetMaximumSize)
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSizeConstraint(QtWidgets.QLayout.SetMaximumSize)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.hLayoutCtrl = QtWidgets.QHBoxLayout()
        self.hLayoutCtrl.setSpacing(1)
        self.hLayoutCtrl.setObjectName("hLayoutCtrl")
        self.loadIpfsButton = QtWidgets.QToolButton(BrowserTabForm)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/share/icons/ipfs-logo-128-ice.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.loadIpfsButton.setIcon(icon)
        self.loadIpfsButton.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.loadIpfsButton.setObjectName("loadIpfsButton")
        self.hLayoutCtrl.addWidget(self.loadIpfsButton)
        self.loadFromClipboardButton = QtWidgets.QToolButton(BrowserTabForm)
        self.loadFromClipboardButton.setText("")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/share/icons/clipboard.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.loadFromClipboardButton.setIcon(icon1)
        self.loadFromClipboardButton.setObjectName("loadFromClipboardButton")
        self.hLayoutCtrl.addWidget(self.loadFromClipboardButton)
        self.horizontalLayout.addLayout(self.hLayoutCtrl)
        self.hashmarkThisPage = QtWidgets.QToolButton(BrowserTabForm)
        self.hashmarkThisPage.setText("")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/share/icons/hashmark-black.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.hashmarkThisPage.setIcon(icon2)
        self.hashmarkThisPage.setObjectName("hashmarkThisPage")
        self.horizontalLayout.addWidget(self.hashmarkThisPage)
        self.urlZone = QtWidgets.QLineEdit(BrowserTabForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.urlZone.sizePolicy().hasHeightForWidth())
        self.urlZone.setSizePolicy(sizePolicy)
        self.urlZone.setMinimumSize(QtCore.QSize(400, 0))
        self.urlZone.setMaximumSize(QtCore.QSize(3000, 26))
        font = QtGui.QFont()
        font.setFamily("Sans Serif")
        font.setPointSize(12)
        font.setItalic(False)
        self.urlZone.setFont(font)
        self.urlZone.setStyleSheet("background-color: rgb(189, 236, 255);")
        self.urlZone.setMaxLength(2048)
        self.urlZone.setObjectName("urlZone")
        self.horizontalLayout.addWidget(self.urlZone)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setSpacing(1)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.stopButton = QtWidgets.QToolButton(BrowserTabForm)
        self.stopButton.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/share/icons/cancel.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.stopButton.setIcon(icon3)
        self.stopButton.setObjectName("stopButton")
        self.horizontalLayout_3.addWidget(self.stopButton)
        self.reloadPageButton = QtWidgets.QToolButton(BrowserTabForm)
        self.reloadPageButton.setText("")
        icon4 = QtGui.QIcon()
        icon4.addPixmap(QtGui.QPixmap(":/share/icons/refresh.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.reloadPageButton.setIcon(icon4)
        self.reloadPageButton.setObjectName("reloadPageButton")
        self.horizontalLayout_3.addWidget(self.reloadPageButton)
        spacerItem = QtWidgets.QSpacerItem(6, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem)
        self.backButton = QtWidgets.QToolButton(BrowserTabForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.backButton.sizePolicy().hasHeightForWidth())
        self.backButton.setSizePolicy(sizePolicy)
        self.backButton.setMaximumSize(QtCore.QSize(32, 16777215))
        self.backButton.setText("")
        icon5 = QtGui.QIcon()
        icon5.addPixmap(QtGui.QPixmap(":/share/icons/left-arrow.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.backButton.setIcon(icon5)
        self.backButton.setObjectName("backButton")
        self.horizontalLayout_3.addWidget(self.backButton)
        self.forwardButton = QtWidgets.QToolButton(BrowserTabForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.forwardButton.sizePolicy().hasHeightForWidth())
        self.forwardButton.setSizePolicy(sizePolicy)
        self.forwardButton.setMaximumSize(QtCore.QSize(32, 16777215))
        self.forwardButton.setText("")
        icon6 = QtGui.QIcon()
        icon6.addPixmap(QtGui.QPixmap(":/share/icons/right-arrow.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.forwardButton.setIcon(icon6)
        self.forwardButton.setObjectName("forwardButton")
        self.horizontalLayout_3.addWidget(self.forwardButton)
        spacerItem1 = QtWidgets.QSpacerItem(6, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem1)
        self.zoomOutButton = QtWidgets.QToolButton(BrowserTabForm)
        self.zoomOutButton.setText("")
        icon7 = QtGui.QIcon()
        icon7.addPixmap(QtGui.QPixmap(":/share/icons/zoom-out.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.zoomOutButton.setIcon(icon7)
        self.zoomOutButton.setObjectName("zoomOutButton")
        self.horizontalLayout_3.addWidget(self.zoomOutButton)
        self.zoomInButton = QtWidgets.QToolButton(BrowserTabForm)
        self.zoomInButton.setText("")
        icon8 = QtGui.QIcon()
        icon8.addPixmap(QtGui.QPixmap(":/share/icons/zoom-in.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.zoomInButton.setIcon(icon8)
        self.zoomInButton.setObjectName("zoomInButton")
        self.horizontalLayout_3.addWidget(self.zoomInButton)
        spacerItem2 = QtWidgets.QSpacerItem(6, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem2)
        self.pinAllButton = QtWidgets.QToolButton(BrowserTabForm)
        self.pinAllButton.setText("")
        icon9 = QtGui.QIcon()
        icon9.addPixmap(QtGui.QPixmap(":/share/icons/pin-black.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pinAllButton.setIcon(icon9)
        self.pinAllButton.setObjectName("pinAllButton")
        self.horizontalLayout_3.addWidget(self.pinAllButton)
        self.pinToolButton = QtWidgets.QToolButton(BrowserTabForm)
        self.pinToolButton.setText("")
        self.pinToolButton.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.pinToolButton.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        self.pinToolButton.setObjectName("pinToolButton")
        self.horizontalLayout_3.addWidget(self.pinToolButton)
        self.horizontalLayout.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_7.addLayout(self.horizontalLayout)
        self.verticalLayout_17.addLayout(self.horizontalLayout_7)
        self.pBarBrowser = QtWidgets.QProgressBar(BrowserTabForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pBarBrowser.sizePolicy().hasHeightForWidth())
        self.pBarBrowser.setSizePolicy(sizePolicy)
        self.pBarBrowser.setMaximumSize(QtCore.QSize(16777215, 2))
        self.pBarBrowser.setProperty("value", 0)
        self.pBarBrowser.setTextVisible(False)
        self.pBarBrowser.setInvertedAppearance(False)
        self.pBarBrowser.setObjectName("pBarBrowser")
        self.verticalLayout_17.addWidget(self.pBarBrowser)
        self.vLayoutBrowser = QtWidgets.QVBoxLayout()
        self.vLayoutBrowser.setSizeConstraint(QtWidgets.QLayout.SetMaximumSize)
        self.vLayoutBrowser.setObjectName("vLayoutBrowser")
        self.verticalLayout_17.addLayout(self.vLayoutBrowser)
        self.verticalLayout_18.addLayout(self.verticalLayout_17)
        self.linkInfosLabel = QtWidgets.QLabel(BrowserTabForm)
        self.linkInfosLabel.setMaximumSize(QtCore.QSize(16777215, 16))
        self.linkInfosLabel.setText("")
        self.linkInfosLabel.setObjectName("linkInfosLabel")
        self.verticalLayout_18.addWidget(self.linkInfosLabel, 0, QtCore.Qt.AlignRight)

        self.retranslateUi(BrowserTabForm)
        QtCore.QMetaObject.connectSlotsByName(BrowserTabForm)

    def retranslateUi(self, BrowserTabForm):
        _translate = QtCore.QCoreApplication.translate
        BrowserTabForm.setWindowTitle(_translate("BrowserTabForm", "Form"))
        self.loadIpfsButton.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Browse IPFS</p></body></html>"))
        self.loadIpfsButton.setText(_translate("BrowserTabForm", "..."))
        self.hashmarkThisPage.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Hashmark this page</p></body></html>"))
        self.stopButton.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Stop</p></body></html>"))
        self.reloadPageButton.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Reload page</p></body></html>"))
        self.backButton.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Previous</p></body></html>"))
        self.forwardButton.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Next</p></body></html>"))
        self.zoomOutButton.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Zoom out</p></body></html>"))
        self.zoomInButton.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Zoom in</p></body></html>"))
        self.pinAllButton.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Automatically pins visited pages</p></body></html>"))

from . import galacteek_rc
